import gnu.trove.set.hash.TCustomHashSet;
import gnu.trove.set.hash.TIntHashSet;
import org.jfree.chart.JFreeChart;

import java.util.HashMap;

public class AnalyticalRepro
{
    public static void main(String[] args)
    {
        double inizio=System.currentTimeMillis();

        //Input parameters
        String netFile=null;
        String queryFile=null;
        int numRandVars=100;
        int minFreq=1;
        double pvalSimulThresh=1.0;
        double pvalAnalThresh=1.0;
        String outputFile=null;
        String outputFilePlot=null;
        String randomFolder=null;
        boolean approx=false;
        boolean append=false;

        int i;
        //Read input parameters
        for (i=0;i<args.length;i++)
        {
            switch (args[i])
            {
                case "-n": netFile=args[++i]; break;
                case "-q": queryFile=args[++i]; break;
                case "-nr": numRandVars=Integer.parseInt(args[++i]); break;
                case "-ps": pvalSimulThresh=Double.parseDouble(args[++i]); break;
                case "-pa": pvalAnalThresh=Double.parseDouble(args[++i]); break;
                case "-f": minFreq=Integer.parseInt(args[++i]); break;
                case "-or": outputFile=args[++i]; break;
                case "-op": outputFilePlot=args[++i]; break;
                case "-rf": randomFolder=args[++i]; break;
                case "-ap": approx=true; break;
                case "-af": append=true; break;
                default: System.out.println("Error! Unrecognizable command '" + args[i] + "'"); printHelp(); System.exit(1);
            }
        }
        //Error in case network file is missing
        if(netFile==null)
        {
            System.out.println("Error! No input network has been specified!\n");
            printHelp();
            System.exit(1);
        }
        if(queryFile==null)
        {
            System.out.println("Error! No query has been specified!\n");
            printHelp();
            System.exit(1);
        }

        //Read input network
        System.out.println("\nReading graph file...");
        FileManager fm=new FileManager();
        Graph net=fm.readGraph(netFile);
        boolean directed=net.isDirected();
        int maxMultiplicity=net.getMaxMultiplicity();
        HashMap<String,TCustomHashSet<TIntHashSet>> condEdgeColors=net.getCondEdgeColors();

        //Read query
        System.out.println("Building motif set...");
        TCustomHashSet<Query> setQueries=fm.readQueries(queryFile,directed,maxMultiplicity);

        System.out.println("-----------------------");
        System.out.println("Searching for colored non-induced injective motifs with color-topology dependence...");

        //Compute motif frequencies
        double startCommon=System.currentTimeMillis();

        ///Build symmetry conditions
        int maxMotifSize=0;
        for(Query q : setQueries)
        {
            q.setMatMachine();
            int motSize=q.getNumNodes();
            if(motSize>maxMotifSize)
                maxMotifSize=motSize;
        }
        MotifManager mm=new MotifManager(net,setQueries);
        TCustomHashSet<Motif> setMotifs=mm.censusGraph(minFreq,condEdgeColors);
        double endCommon=System.currentTimeMillis();
        double totalTimeCommon=(endCommon-startCommon)/1000;


        double timeSimul=0.0;
        double timeAnal=0.0;
        if(maxMotifSize>0)
        {
            //Compute motif frequencies in a sample of EDD random networks
            double startSimul=System.currentTimeMillis();
            System.out.println("Setting up EDD model...");
            EDDModel edd=new EDDModel(net,maxMotifSize,approx);
            mm.censusRandom(edd,numRandVars,setMotifs,pvalSimulThresh,randomFolder,fm);
            double endSimul=System.currentTimeMillis();
            timeSimul=(endSimul-startSimul)/1000;

            //Compute analytical p-values
            double startAnal=System.currentTimeMillis();
            mm.computeAnalyticalPVals(edd,setMotifs,pvalAnalThresh,approx);
            double endAnal=System.currentTimeMillis();
            timeAnal=(endAnal-startAnal)/1000;
        }

        //Print results
        if(outputFile!=null)
        {
            fm.writeResults(outputFile,setMotifs,directed,true,append);
            System.out.println("Results written to "+outputFile);
        }
        else
            fm.printResults(setMotifs,directed,true);


        //Create Precision-recall plot and write it to output file
        if(outputFilePlot!=null)
        {
            System.out.println("Making precision-recall plot... ");
            PlotManager pm=new PlotManager(netFile,maxMotifSize);
            double thresholdPVal=0.05;
            JFreeChart chart=pm.createPRPlot(setMotifs,thresholdPVal);
            fm.savePlotToPDFFile(chart,outputFilePlot);
            System.out.println("Done! Plot saved in "+outputFilePlot);
        }

        System.out.println("Running time for estimating motif frequencies: "+totalTimeCommon+" secs");
        System.out.println("Running time of analytical p-values computation: "+timeAnal+" secs");
        System.out.println("Running time of simulation p-values computation: "+timeSimul+" secs");
        double fine=System.currentTimeMillis();
        double totalTime=(fine-inizio)/1000;
        System.out.println("\nTotal time elapsed: "+totalTime+" secs");

    }

    private static void printHelp()
    {
        String help = "Usage: java -jar AnalyticalRepro.jar -n <networkFile> -q <queryFile> "+
                "[-nr <numRandomVariants> -ps <pvalSimulThresh> -pa <pvalAnalThresh> "+
                "-f <minFrequency> -or <resultsFile> -op <plotFile> -rf <randomFolder> -ap -af]\n\n";
        help+="REQUIRED PARAMETERS:\n";
        help+="-n\tInput network file\n";
        help+="-q\tQuery file\n\n";
        help+="OPTIONAL PARAMETERS:\n";
        help+="-nr\tNumber of random variants used for computing simulation-based p-value (default=100)\n";
        help+="-ps\tReturn only motifs with simulation-based p-value below this threshold (default=1.0)\n";
        help+="-pa\tReturn only motifs with analytical-based p-value below this threshold (default=1.0)\n";
        help+="-f\tReturn only motifs whose frequency in the input network is greater than or equals to this value (default=1)\n";
        help+="-or\tOutput file where results will be saved (default=results.txt)\n";
        help+="-op\tOutput file where Precision-Recall plot will be saved (default=prPlot.pdf)\n";
        help+="-rf\tRandom folder with prevoiusly built random graphs\n";
        help+="-ap\tCompute approximate p-values\n";
        help+="-af\tAppend results to the output file\n";
        System.out.println(help);
    }
}
