import gnu.trove.iterator.TIntIterator;
import gnu.trove.map.custom_hash.TObjectIntCustomHashMap;
import gnu.trove.map.hash.TIntDoubleHashMap;
import gnu.trove.map.hash.TIntObjectHashMap;
import gnu.trove.set.hash.TCustomHashSet;
import gnu.trove.set.hash.TIntHashSet;

import java.util.*;

public class EDDModel
{
	private int numNodes;
	private boolean directed;
	private double[] freqColors; //Frequencies of colors in the network
	private double[] gamma;
	private TObjectIntCustomHashMap<TIntHashSet> mapEdgeColsIds;
	private TIntDoubleHashMap[][] outDistrDep;
	private TIntDoubleHashMap[][] inDistrDep;
	private double[][][] momentsOutDep;
	private double[][][] momentsInDep;
	public EDDModel(Graph net, int maxMotifSize, boolean approx)
	{
		this.numNodes=net.getNumNodes();
		this.directed=net.isDirected();
		this.freqColors=net.getFreqNodeColors();
		this.mapEdgeColsIds=net.getMapEdgeColsIds();
		this.gamma=new double[mapEdgeColsIds.size()];
		TIntDoubleHashMap[][][] multiColorDegDistr=net.getMultiColorDegDistr();
		this.outDistrDep=multiColorDegDistr[0];
		this.inDistrDep=multiColorDegDistr[1];
		computeGamma();
		if(approx)
			setMoments(maxMotifSize-1);
		else
			setMoments(2*maxMotifSize-2);
	}

	public Graph sampleGraph(Graph net, LinkedList<Integer>[][][] listDegrees)
	{
		Random r=new Random();
		int i,k;
		Vector<TIntHashSet> listColEdges=net.getListEdgeCols();

		//Create graph
		Graph randNet=new Graph(net);

		//Sample node labels
		int[] colorNodeNets=net.getNodeListColors();
		int[] labelNodeRand=new int[colorNodeNets.length];
		for(i=0;i<numNodes;i++)
			labelNodeRand[i]=colorNodeNets[i];

		//Sample node degrees given node and edge labels
		LinkedList<Integer>[][] degreeWeights=GraphUtility.assignDegrees(labelNodeRand,listDegrees,directed);

		//Add nodes
		for(i=0;i<numNodes;i++)
			randNet.addNode(i,labelNodeRand[i]);

		//Add edges
		LinkedList<Integer> queuePosOut=new LinkedList<>();
		LinkedList<Integer> queuePosIn=new LinkedList<>();
		for(i=0;i<degreeWeights[0].length;i++)
		{
			LinkedList<Integer> listDegsOut=degreeWeights[0][i];
			LinkedList<Integer> listDegsIn=degreeWeights[1][i];
			int numEdgesMulti=listDegsOut.size();
			if(listDegsIn.size()==0)
				numEdgesMulti=0;
			if(!directed)
				numEdgesMulti/=2;
			k=0;
			int attempt=0;
			while(k<numEdgesMulti)
			{
				int sourceNode;
				int destNode;
				if(queuePosOut.isEmpty())
				{
					sourceNode=listDegsOut.get(r.nextInt(listDegsOut.size()));
					if(queuePosIn.isEmpty())
						destNode=listDegsIn.get(r.nextInt(listDegsIn.size()));
					else
						destNode=queuePosIn.removeFirst();
				}
				else
				{
					if(queuePosIn.isEmpty())
					{
						sourceNode=queuePosOut.removeFirst();
						destNode=listDegsIn.get(r.nextInt(listDegsIn.size()));
					}
					else
					{
						if(r.nextDouble()<0.5)
						{
							sourceNode=queuePosOut.removeFirst();
							destNode=listDegsIn.get(r.nextInt(listDegsIn.size()));
						}
						else
						{
							sourceNode=listDegsOut.get(r.nextInt(listDegsOut.size()));
							destNode=queuePosIn.removeFirst();
						}
					}
				}
				if(sourceNode!=destNode && !randNet.isEdge(sourceNode,destNode))
				{
					TIntHashSet colsEdge=listColEdges.get(i);
					randNet.addEdge(sourceNode,destNode,colsEdge);
					k++;
					attempt=0;
				}
				else
				{
					queuePosOut.addLast(sourceNode);
					queuePosIn.addLast(destNode);
					attempt++;
				}
				if(attempt==1000)
					break;
			}
		}

		return randNet;
	}

	protected void computeGamma()
	{
		int i, j;
		for(j=0;j<outDistrDep[0].length;j++)
		{
			double meanDegree=0.0;
			for(i=0;i<outDistrDep.length;i++)
			{
				TIntIterator it=outDistrDep[i][j].keySet().iterator();
				while(it.hasNext())
				{
					int deg=it.next();
					double valDeg=outDistrDep[i][j].get(deg);
					meanDegree+=valDeg*freqColors[i]*deg;
				}
			}
			if(meanDegree!=0)
				gamma[j]=1.0/((numNodes-1)*meanDegree);
			else
				gamma[j]=1.0;
		}
	}

	protected double[] getCoeffOverlap(int k)
	{
		double[] coeffOverlap=new double[k+1];
		int[] paramMulti=new int[3];
		paramMulti[0]=numNodes-2*k;
		paramMulti[1]=paramMulti[2]=k;
		coeffOverlap[0] = MathUtility.multinomCoeff(numNodes,paramMulti);
		paramMulti=new int[4];
		int s;
		for(s=1;s<=k;s++)
		{
			paramMulti[0]=k-s;
			paramMulti[1]=s;
			paramMulti[2]=k-s;
			paramMulti[3]=numNodes-2*k+s;
			coeffOverlap[s]=MathUtility.multinomCoeff(numNodes,paramMulti);
		}
		return coeffOverlap;
	}

	double computeMean(Motif mot, double numNonRedPerm)
	{
		double probOcc=computeOccProb(mot);
		double binomCoeff=MathUtility.binomCoeff(numNodes,mot.getNodeLabels().length);
		//System.out.println("binomCoeff: "+binomCoeff);
		//System.out.println("probOcc: "+probOcc);
		//System.out.println("numNonRedPerm: "+numNonRedPerm);
		return binomCoeff*numNonRedPerm*probOcc;
		//System.out.println("mean: "+mean);
	}

	double computeVariance(Motif mot, TCustomHashSet<Motif> nrPerm, double mean)
	{
		int l, m;

		int k=mot.getNumNodes();
		double[] coeffOverlap=getCoeffOverlap(k);

		// Compute first part of second moment
		int numberColPerms=nrPerm.size();
		double varPerm = numberColPerms*computeOccProb(mot);
		double moment2 = coeffOverlap[0]*varPerm*varPerm;

		//Compute second part of second moment
		double overlappingSum=0.0;
		for(int s=1;s<=k;s++)
		{
			double probSuperMotif=0.0;
			for(Motif mSource : nrPerm)
			{
				TIntObjectHashMap[] edgeColsSource=mSource.getOutAdjLists();
				int[] nodeColsSource=mSource.getNodeLabels();
				for(Motif mDest : nrPerm)
				{
					TIntObjectHashMap[] edgeColsDest=mDest.getOutAdjLists();
					int[] nodeColsDest=mDest.getNodeLabels();

					//Check if colors in overlap region are the same
					boolean diffColorsOverlap=false;
					for(l=0;l<s;l++)
					{
						if(nodeColsSource[k-s+l]!=nodeColsDest[l])
						{
							diffColorsOverlap=true;
							break;
						}
					}
					if(!diffColorsOverlap)
					{
						boolean diffEdgeOverlap=false;
						for(l=k-s;l<k;l++)
						{
							for (m=k-s;m<k;m++)
							{
								TIntHashSet edgeSource=(TIntHashSet) edgeColsSource[l].get(m);
								TIntHashSet edgeDest=(TIntHashSet) edgeColsDest[l-k+s].get(m-k+s);
								if((edgeSource==null && edgeDest!=null) ||
								(edgeSource!=null && edgeDest==null) ||
								(edgeSource!=null && !edgeSource.equals(edgeDest)))
								{
									diffEdgeOverlap=true;
									break;
								}
							}
							if(diffEdgeOverlap)
								break;
						}
						if(!diffEdgeOverlap)
						{
							Motif mSuper=GraphUtility.createSuperMotif(mSource,mDest,s);
							double occProbSuper=computeOccProb(mSuper);
							probSuperMotif+=occProbSuper;
						}
					}
				}
			}
			overlappingSum+=coeffOverlap[s]*probSuperMotif;
		}
		moment2+=overlappingSum;

		//Compute variance
		return moment2-(mean*mean);
	}

	private double computeOccProb(Motif mot)
	{
		double probQuery=1.0;
		int[] nodeLabs=mot.getNodeLabels();
		TIntObjectHashMap[] edgeLabs=mot.getOutAdjLists();
		int[][] degreeOccOut=new int[nodeLabs.length][gamma.length];
		int[][] degreeOccIn=new int[nodeLabs.length][gamma.length];
		int i, j;

		//Compute node degrees
		for(i=0;i<edgeLabs.length;i++)
		{
			TIntIterator it=edgeLabs[i].keySet().iterator();
			while(it.hasNext())
			{
				j=it.next();
				int posCol=mapEdgeColsIds.get(edgeLabs[i].get(j));
				degreeOccOut[i][posCol]++;
				degreeOccIn[j][posCol]++;
				if(directed || i<j)
					probQuery*=gamma[posCol];
			}
		}

		//Prob moments
		for(i=0;i<degreeOccOut.length;i++)
		{
			probQuery*=freqColors[nodeLabs[i]];
			for(j=0;j<gamma.length;j++)
			{
				probQuery*=momentsOutDep[nodeLabs[i]][j][degreeOccOut[i][j]];
				if(directed)
					probQuery*=momentsInDep[nodeLabs[i]][j][degreeOccIn[i][j]];
			}
		}

		return probQuery;
	}

	protected void setMoments(int degree)
	{
		momentsOutDep=new double[outDistrDep.length][gamma.length][degree+1];
		momentsInDep=new double[inDistrDep.length][gamma.length][degree+1];
		int i, j, k;
		for(i=0;i<momentsOutDep.length;i++)
		{
			for(j=0;j<momentsOutDep[i].length;j++)
			{
				momentsOutDep[i][j][0]=1.0;
				momentsInDep[i][j][0]=1.0;
			}
			for(j=0;j<momentsOutDep[i].length;j++)
			{
				for(k=0;k<momentsOutDep[i][j].length;k++)
				{
					momentsOutDep[i][j][k]=0.0;
					momentsInDep[i][j][k]=0.0;
					TIntIterator it=outDistrDep[i][j].keySet().iterator();
					while(it.hasNext())
					{
						int deg=it.next();
						double valDeg=outDistrDep[i][j].get(deg);
						momentsOutDep[i][j][k]+=valDeg*Math.pow(deg,k);
					}
					it=inDistrDep[i][j].keySet().iterator();
					while(it.hasNext())
					{
						int deg=it.next();
						double valDeg=inDistrDep[i][j].get(deg);
						momentsInDep[i][j][k]+=valDeg*Math.pow(deg,k);
					}
				}
			}
		}
	}
}
	