import com.itextpdf.awt.DefaultFontMapper;
import com.itextpdf.awt.FontMapper;
import com.itextpdf.text.Document;
import com.itextpdf.text.DocumentException;
import com.itextpdf.text.Rectangle;
import com.itextpdf.text.pdf.PdfContentByte;
import com.itextpdf.text.pdf.PdfTemplate;
import com.itextpdf.text.pdf.PdfWriter;

import java.awt.*;
import java.awt.geom.Rectangle2D;
import gnu.trove.iterator.TIntIterator;
import gnu.trove.map.hash.TIntObjectHashMap;
import gnu.trove.set.hash.TCustomHashSet;
import gnu.trove.set.hash.TIntHashSet;
import org.jfree.chart.JFreeChart;

import java.io.*;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Vector;

public class FileManager
{
    private HashMap<String,Integer> mapNodeLabels;
    private HashMap<String,Integer> mapEdgeLabels;
    private Vector<String> setNodeLabels;
    private Vector<String> setEdgeLabels;

    public FileManager()
    {
        this.mapNodeLabels=new HashMap<>();
        this.setNodeLabels=new Vector<>();
        this.mapEdgeLabels=new HashMap<>();
        this.setEdgeLabels=new Vector<>();
    }

    public Graph readGraph(String graphFile)
    {
        Graph g=null;
        try
        {
            BufferedReader br=new BufferedReader(new FileReader(graphFile));
            String str=br.readLine();
            boolean directed=false;
            if(str.equals("directed"))
                directed=true;
            int numNodes=Integer.parseInt(br.readLine());
            g=new Graph(directed,numNodes);
            String[] split=graphFile.split("[\\\\|/]");
            String[] split2=split[split.length-1].split("\\.");
            g.setName(split2[0]);

            //Read node colors
            int i;
            for(i=0;i<numNodes;i++)
            {
                String lab = br.readLine();
                if(!mapNodeLabels.containsKey(lab))
                {
                    mapNodeLabels.put(lab, setNodeLabels.size());
                    setNodeLabels.add(lab);
                }
                int nodeCol=mapNodeLabels.get(lab);
                g.addNode(i,nodeCol);
            }

            //Read edges
            while((str=br.readLine())!=null)
            {
                split=str.split("\t");
                int source=Integer.parseInt(split[0]);
                int dest=Integer.parseInt(split[1]);
                split2=split[2].split(",");
                for(i=0;i<split2.length;i++)
                {
                    if(!mapEdgeLabels.containsKey(split2[i]))
                    {
                        mapEdgeLabels.put(split2[i],setEdgeLabels.size());
                        setEdgeLabels.add(split2[i]);
                    }
                    int edgeCol=mapEdgeLabels.get(split2[i]);
                    g.addEdge(source,dest,edgeCol);
                }
            }
            br.close();
            g.setListEdgeCols();
            g.setNumEdgeCols(setEdgeLabels.size());
        }
        catch(Exception e){
            System.out.println(e.getMessage());
        }
        return g;
    }

    TCustomHashSet<Query> readQueries(String queriesFile, boolean directed, int maxMultiplicity)
    {
        TCustomHashSet<Query> setQueries=new TCustomHashSet<>(new QueryStrategy());
        try
        {
            BufferedReader br=new BufferedReader(new FileReader(queriesFile));
            String str=br.readLine();
            while(str!=null)
            {
                int numNodes=Integer.parseInt(br.readLine());
                Query q=new Query(numNodes,directed);
                int i;
                boolean present=true;
                //Read node labels
                for(i=0;i<numNodes;i++)
                {
                    String lab = br.readLine();
                    if(!lab.equals("?"))
                    {
                        if(!mapNodeLabels.containsKey(lab))
                        {
                            present=false;
                            break;
                        }
                        else
                        {
                            int nodeCol=mapNodeLabels.get(lab);
                            q.addNode(i,nodeCol);
                        }
                    }
                    else
                        q.addNode(i,-1);
                }

                if(present)
                {
                    //Read edges
                    while((str=br.readLine())!=null && !str.startsWith("#"))
                    {
                        String[] split=str.split("\t");
                        int source=Integer.parseInt(split[0]);
                        int dest=Integer.parseInt(split[1]);
                        String[] split2=split[2].split(",");
                        TIntHashSet setEdges=new TIntHashSet();
                        int minMulti=split2.length;
                        int maxMulti=minMulti;
                        for (i=0;i<split2.length;i++)
                        {
                            if(!split2[i].equals("?") && !split2[i].equals("*"))
                            {
                                if(!mapEdgeLabels.containsKey(split2[i]))
                                {
                                    present=false;
                                    break;
                                }
                                else
                                {
                                    int edgeCol=mapEdgeLabels.get(split2[i]);
                                    setEdges.add(edgeCol);
                                }
                            }
                            else if(split2[i].equals("*"))
                                maxMulti=maxMultiplicity;
                        }
                        if(!present)
                            break;
                        q.addEdge(source,dest,setEdges,minMulti,maxMulti);
                    }
                    if(present)
                        setQueries.add(q);
                    else
                    {
                        str=br.readLine();
                        while(str!=null && !str.startsWith("#"))
                            str=br.readLine();
                    }
                }
            }
            br.close();
        }
        catch(Exception e){
            System.out.println(e.toString());
        }
        return setQueries;
    }

    void writeGraph(Graph net, String graphFile)
    {
        try
        {
            BufferedWriter bw=new BufferedWriter(new FileWriter(graphFile));
            boolean directed=net.isDirected();
            if(directed)
                bw.write("directed\n");
            else
                bw.write("undirected\n");
            Vector<String> setNodeLabels=getSetNodeLabels();
            Vector<String> setEdgeLabels=getSetEdgeLabels();
            int[] nodeColors=net.getNodeListColors();
            int i;
            bw.write(nodeColors.length+"\n");
            for(i=0;i<nodeColors.length;i++)
                bw.write(setNodeLabels.get(nodeColors[i])+"\n");
            TIntObjectHashMap[] outAdjList=net.getOutAdjList();
            for(i=0;i<outAdjList.length;i++)
            {
                TIntIterator it=outAdjList[i].keySet().iterator();
                while(it.hasNext())
                {
                    int dest=it.next();
                    if(directed || i<dest)
                    {
                        bw.write(i + "\t" + dest);
                        TIntHashSet setEdgeCols = (TIntHashSet) outAdjList[i].get(dest);
                        TIntIterator itCols=setEdgeCols.iterator();
                        int lab=itCols.next();
                        bw.write("\t" + setEdgeLabels.get(lab));
                        while(itCols.hasNext())
                        {
                            lab=itCols.next();
                            bw.write("," + setEdgeLabels.get(lab));
                        }
                        bw.write("\n");
                    }
                }
            }
            bw.close();
        }
        catch(Exception e){
            System.out.println(e.getMessage());
        }
    }

    void writeResults(String outputFile, TCustomHashSet<Motif> setMotifs, boolean directed, boolean extended, boolean append)
    {
        int j, k;
        try
        {
            String header="";
            if(!append)
            {
                header+="Motif_Nodes\tMotif_edges\tNum_occ_input_graph\tMean_analytical\tVariance_analytical\tPval_analytical";
                if(extended)
                    header+="\tMean_simulation\tPval_simulation";
            }
            BufferedWriter bw=new BufferedWriter(new FileWriter(outputFile,append));
            if(!header.equals(""))
                bw.write(header+"\n");
            Iterator<Motif> it=setMotifs.iterator();
            long numTotalOccs=0;
            while(it.hasNext())
            {
                Motif occ=it.next();
                long numOccs=occ.getNumOccurrences();
                int[] nodeColors=occ.getNodeLabels();
                TIntObjectHashMap[] adjLists=occ.getOutAdjLists();
                for(j=0;j<nodeColors.length-1;j++)
                    bw.write(setNodeLabels.get(nodeColors[j])+",");
                bw.write(setNodeLabels.get(nodeColors[j])+"\t");
                for(j=0;j<adjLists.length;j++)
                {
                    TIntIterator itEdge=adjLists[j].keySet().iterator();
                    while(itEdge.hasNext())
                    {
                        k=itEdge.next();
                        if(directed || j<k)
                        {
                            bw.write("("+j+","+k);
                            TIntHashSet edge=(TIntHashSet) adjLists[j].get(k);
                            TIntIterator itEdgeCols=edge.iterator();
                            while(itEdgeCols.hasNext())
                            {
                                int edgeCol=itEdgeCols.next();
                                String edgeLab=setEdgeLabels.get(edgeCol);
                                bw.write(","+edgeLab);
                            }
                            bw.write(")");
                        }
                    }
                }
                bw.write("\t"+numOccs+"\t"+occ.getAvgAnal()+"\t"+occ.getVarianceAnal()+"\t"+occ.getPValAnal());
                if(extended)
                    bw.write("\t"+occ.getAvgRandom()+"\t"+occ.getPValSimul());
                bw.write("\n");
                numTotalOccs+=numOccs;
            }
            System.out.println("Found "+numTotalOccs+" occurrences and "+setMotifs.size()+" motifs.");
            bw.close();
        }
        catch(Exception e){
            System.out.println(e.getMessage());
        }
    }

    public Vector<String> getSetNodeLabels()
    {
        return setNodeLabels;
    }

    public Vector<String> getSetEdgeLabels()
    {
        return setEdgeLabels;
    }

    void savePlotToPDFFile(JFreeChart chart, String pdfFile)
    {
        try
        {
            File f=new File(pdfFile);
            int width=800;
            int height=500;
            saveChartAsPDF(f,chart,width,height,new DefaultFontMapper());
        }
        catch(Exception e){
            System.out.println(e.getMessage());
        }

    }
    private void writeChartAsPDF(OutputStream out, JFreeChart chart, int width, int height, FontMapper mapper)
    {
        Rectangle pagesize=new Rectangle(width,height);
        Document document=new Document(pagesize,50,50,50,50);
        try
        {
            PdfWriter writer=PdfWriter.getInstance(document,out);
            document.addAuthor("Micale");
            document.addSubject("Prova");
            document.open();
            PdfContentByte cb=writer.getDirectContent();
            PdfTemplate tp=cb.createTemplate(width,height);
            Graphics2D g2=tp.createGraphics(width,height,mapper);
            Rectangle2D r2D=new Rectangle2D.Double(0,0,width,height);
            chart.draw(g2,r2D);
            g2.dispose();
            cb.addTemplate(tp,0,0);
        }
        catch(DocumentException e){
            System.out.println(e.getMessage());
        }
        document.close();
    }

    private void saveChartAsPDF(File file, JFreeChart chart, int width, int height, FontMapper mapper)
    {
        try
        {
            OutputStream out = new BufferedOutputStream(new FileOutputStream(file));
            writeChartAsPDF(out, chart, width, height, mapper);
            out.close();
        }
        catch(Exception e) {
            System.out.println(e.toString());
        }
    }

    void printResults(TCustomHashSet<Motif> setMotifs, boolean directed, boolean extended)
    {
        int j, k;
        try
        {
            String header="Motif_Nodes\tMotif_edges\tNum_occ_input_graph\tMean_analytical\tVariance_analytical\tPval_analytical";
            if(extended)
                header+="\tMean_simulation\tPval_simulation";
            System.out.println(header);
            Iterator<Motif> it=setMotifs.iterator();
            long numTotalOccs=0;
            while(it.hasNext())
            {
                Motif occ=it.next();
                long numOccs=occ.getNumOccurrences();
                int[] nodeColors=occ.getNodeLabels();
                TIntObjectHashMap[] adjLists=occ.getOutAdjLists();
                for(j=0;j<nodeColors.length-1;j++)
                    System.out.print(setNodeLabels.get(nodeColors[j])+",");
                System.out.print(setNodeLabels.get(nodeColors[j])+"\t");
                for(j=0;j<adjLists.length;j++)
                {
                    //TIntObjectHashMap<TIntHashSet> adjList=adjLists[j];
                    TIntIterator itEdge=adjLists[j].keySet().iterator();
                    while(itEdge.hasNext())
                    {
                        k=itEdge.next();
                        if(directed || j<k)
                        {
                            System.out.print("("+j+","+k);
                            TIntHashSet edge=(TIntHashSet) adjLists[j].get(k);
                            TIntIterator itEdgeCol=edge.iterator();
                            while(itEdgeCol.hasNext())
                            {
                                int edgeCol=itEdgeCol.next();
                                String edgeLab=setEdgeLabels.get(edgeCol);
                                System.out.print(","+edgeLab);
                            }
                            System.out.print(")");
                        }
                    }
                }
                System.out.println("\t"+numOccs+"\t"+occ.getAvgAnal()+"\t"+occ.getVarianceAnal()+"\t"+occ.getPValAnal()+"\t"+occ.getAvgRandom()+"\t"+occ.getPValSimul());
                numTotalOccs+=numOccs;
            }
            System.out.println("Found "+numTotalOccs+" occurrences and "+setMotifs.size()+" motifs.");
        }
        catch(Exception e){
            System.out.println(e.getMessage());
        }
    }
}
