import gnu.trove.set.hash.TCustomHashSet;
import gnu.trove.set.hash.TIntHashSet;

import java.util.HashMap;

public class MultiMotif
{
    public static void main(String[] args)
    {

        //Input parameters
        String netFile=null;
        String queryFile=null;
        int minFreq=1;
        double pvalAnalThresh=1.0;
        String outputFile=null;
        boolean approx=false;
        boolean append=false;

        int i;
        //Read input parameters
        for (i=0;i<args.length;i++)
        {
            switch (args[i])
            {
                case "-n": netFile=args[++i]; break;
                case "-q": queryFile=args[++i]; break;
                case "-pa": pvalAnalThresh=Double.parseDouble(args[++i]); break;
                case "-f": minFreq=Integer.parseInt(args[++i]); break;
                case "-or": outputFile=args[++i]; break;
                case "-ap": approx=true; break;
                case "-af": append=true; break;
                default: System.out.println("Error! Unrecognizable command '" + args[i] + "'"); printHelp(); System.exit(1);
            }
        }
        //Error in case network file is missing
        if(netFile==null)
        {
            System.out.println("Error! No input network has been specified!\n");
            printHelp();
            System.exit(1);
        }
        if(queryFile==null)
        {
            System.out.println("Error! No query has been specified!\n");
            printHelp();
            System.exit(1);
        }

        //Read input network
        System.out.println("\nReading graph file...");
        FileManager fm=new FileManager();
        Graph net=fm.readGraph(netFile);
        boolean directed=net.isDirected();
        int maxMultiplicity=net.getMaxMultiplicity();
        HashMap<String,TCustomHashSet<TIntHashSet>> condEdgeColors=net.getCondEdgeColors();

        //Read query
        System.out.println("Building motif set...");
        TCustomHashSet<Query> setQueries=fm.readQueries(queryFile,directed,maxMultiplicity);

        System.out.println("-----------------------");
        System.out.println("Searching for colored non-induced injective motifs with color-topology dependence...");

        MotifManager mm=new MotifManager(net,setQueries);

        //Compute motif frequencies
        double startInput=System.currentTimeMillis();
        int maxMotifSize=0;
        for(Query q : setQueries)
        {
            q.setMatMachine();
            int motSize=q.getNumNodes();
            if(motSize>maxMotifSize)
                maxMotifSize=motSize;
        }
        TCustomHashSet<Motif> setMotifs=mm.censusGraph(minFreq,condEdgeColors);
        double endInput=System.currentTimeMillis();

        //Compute analytical p-values
        double startAnal=System.currentTimeMillis();
        EDDModel edd=new EDDModel(net,maxMotifSize,approx);
        mm.computeAnalyticalPVals(edd,setMotifs,pvalAnalThresh,approx);
        double endAnal=System.currentTimeMillis();

        //Print results
        if(outputFile!=null)
        {
            fm.writeResults(outputFile,setMotifs,directed,false,append);
            System.out.println("Results written to "+outputFile);
        }
        else
            fm.printResults(setMotifs,directed,false);

        double totalTimeInput=(endInput-startInput)/1000;
        double totalTimeAnal=(endAnal-startAnal)/1000;
        System.out.println("\nRunning time for estimating motif frequencies: "+totalTimeInput+" secs");
        System.out.println("Running time of analytical p-values computation: "+totalTimeAnal+" secs");
    }

    private static void printHelp()
    {
        String help = "Usage: java -jar MultiMotif.jar -n <networkFile> -q <queryFile> "+
                "[-pa <pvalAnalThresh> -f <minFrequency> -or <resultsFile> -ap]\n\n";
        help+="REQUIRED PARAMETERS:\n";
        help+="-n\tInput network file\n";
        help+="-q\tQuery file\n\n";
        help+="OPTIONAL PARAMETERS:\n";
        help+="-pa\tReturn only motifs with analytical-based p-value below this threshold (default=1.0)\n";
        help+="-f\tReturn only motifs whose frequency in the input network is greater than or equals to this value (default=1)\n";
        help+="-or\tOutput file where results will be saved (default=results.txt)\n";
        help+="-ap\tCompute approximate p-values\n";
        //help+="-af\tAppend results to the output file\n";
        System.out.println(help);
    }
}
